unit TMatrix;
// Direct3D Immediate mode test by Massimo Bardetti - bigben@dircon.co.uk.

// This unit implements TMatrix, TLight and TMaterial.
interface
{$A+} {$Z4}
uses Forms, SysUtils,  OLE2, Windows , DDraw, D3D, D3dTypes;

// T3DMatrix :
// T3DMatrix is a simple wrapper to the D3DMATRIX record
// setting the _xy values in a TMatrix object sets the corresponding value in the D3DMATRIX record
// Once you set values remember to call Update to refresh the internal (not in this class but
// somewhere into the D3DDevice object) representation of the matrix
//
// Create : to create a matrix you need a Directx3d device
// GetHandle : returns the handle to the matrix;
// Update : Loads the matrix into the device

type
  T3dMatrix = class ( Tobject )
  private
         m_pID3D : IDirect3dDevice;
         m_hMatrix : D3DMATRIXHANDLE;
         m_mat : D3DMATRIX;
  public


    constructor Create ( pID3D : IDirect3dDevice ) ;
    //destructor Destroy ; override ;
    function GetHandle : D3DMATRIXHANDLE;
    procedure Update;
    property _11 : D3DVALUE read m_mat._11 write m_mat._11;
    property _21 : D3DVALUE read m_mat._21 write m_mat._21;
    property _31 : D3DVALUE read m_mat._31 write m_mat._31;
    property _41 : D3DVALUE read m_mat._41 write m_mat._41;

    property _12 : D3DVALUE read m_mat._12 write m_mat._12;
    property _22 : D3DVALUE read m_mat._22 write m_mat._22;
    property _32 : D3DVALUE read m_mat._32 write m_mat._32;
    property _42 : D3DVALUE read m_mat._42 write m_mat._42;

    property _13 : D3DVALUE read m_mat._13 write m_mat._13;
    property _23 : D3DVALUE read m_mat._23 write m_mat._23;
    property _33 : D3DVALUE read m_mat._33 write m_mat._33;
    property _43 : D3DVALUE read m_mat._43 write m_mat._43;

    property _14 : D3DVALUE read m_mat._14 write m_mat._14;
    property _24 : D3DVALUE read m_mat._24 write m_mat._24;
    property _34 : D3DVALUE read m_mat._34 write m_mat._34;
    property _44 : D3DVALUE read m_mat._44 write m_mat._44;

  end ;

// T3DLIGHT :
// T3DLight implements the iDirect3DLight interface
// Same as the TMatrix you need to call update fater you set any value
//
// Create : creates a light object , needs a IDirect3D interface.
// The other functions are obvious.

type
    T3dLIGHT = class (TObject)
    private
      m_Light : D3DLIGHT;
      m_pILight : IDirect3dLight;
    public
      constructor Create(pIEngine : IDirect3d);

      function GetInterface : IDirect3dLight;
      procedure SetType(ltype : D3DLIGHTTYPE);
      procedure SetColor(r,g,b : double);
      procedure SetDirection(x,y,z : double);
      procedure SetPosition(x,y,z : double);

      function Update : boolean;
end;


// T3DMaterial:
// T3dMaterial implements the iDirect3DMaterial interface
// Now this is auto updating, every time you set a value it will refresh the internal image
// The main reason this is different is because the original C++ classes where done this way
// But I am almost sure now that there is no reason for the update function used in the
// two previous classes to exist, they should also be auto-updating.

type
    T3dMaterial = class (TObject)
    private
      m_Material : D3DMATERIAL;
      m_pIMat : IDirect3dMaterial;
      m_hMat : D3DMATERIALHANDLE;
    public
      constructor Create(pIEngine : IDirect3d; pIDevice : IDirect3dDevice);

      function GetInterface : IDirect3dMaterial;
      function GetHandle : D3DMATERIALHANDLE;

// SetShades : This value should be 1 for backgrounds
      function SetShades(n : integer) : boolean;
      function SetColor(r,g,b : double) : boolean;
      function SetTextureHandle(h : integer) : boolean;
end;


implementation

var
   DDResult : HRESULT;

constructor T3dMatrix.Create(pID3D : IDirect3DDevice);
begin
     m_mat._11 := 1.0; m_mat._12 := 0.0; m_mat._13 := 0.0; m_mat._14 := 0.0;
     m_mat._21 := 0.0; m_mat._22 := 1.0; m_mat._23 := 0.0; m_mat._24 := 0.0;
     m_mat._31 := 0.0; m_mat._32 := 0.0; m_mat._33 := 1.0; m_mat._34 := 0.0;
     m_mat._41 := 0.0; m_mat._42 := 0.0; m_mat._43 := 0.0; m_mat._44 := 1.0;


     m_pID3D := nil;
     m_hMatrix := 0;

     if (pID3D <> nil) then
        begin
             m_pID3D := pID3D;
             m_pID3D.AddRef;

             DDResult := m_pID3D.CreateMatrix(m_hMatrix);
             if (DDResult <> DD_OK) then
                raise Exception.Create (
                       Format ( 'CreateMatrix failed: %x', [ DDResult ] ) ) ;

             Update;
        end;

end;

function T3dMatrix.GetHandle : D3DMATRIXHANDLE;
begin
     result := m_hMatrix;
end;

procedure T3dMatrix.Update;
begin
     if (m_hMatrix <> 0) and (m_pID3D <> nil) then
        begin
             DDResult := m_pID3D.SetMatrix(m_hMatrix, m_mat);
             if (DDResult <> DD_OK) then
                 raise Exception.Create (
                       Format ( 'SetMatrix failed: %x', [ DDResult ] ) );
        end;
end;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

constructor T3DLight.Create(pIEngine : IDirect3d);
begin
     ZeroMemory(@m_Light,sizeof(D3DLIGHT));
     m_light.dwSize := sizeof(d3DLight);
     m_light.dltType := D3dLIGHT_DIRECTIONAL;
     
     DdResult := pIEngine.CreateLight(m_pILight, nil);

     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create('Cannot create light!!');
end;

function T3DLight.GetInterface : IDirect3dLight;
begin
     Result := m_pILight;
end;

procedure T3DLight.SetType(ltype : D3DLIGHTTYPE);
begin
     m_Light.dltType := ltype;
end;

procedure T3dLight.SetColor(r,g,b : double);
var
   c : D3DCOLORVALUE;
begin
     c.r := r;
     c.g := g;
     c.b := b;
     c.a := 1;
     m_Light.dcvColor := c;
end;

procedure T3dLight.SetDirection(x,y,z : double);
var
   v : D3DVECTOR;
begin
     v.x := x;
     v.y := y;
     v.z := z;
     m_Light.dvDirection := v;
end;

procedure T3dLight.SetPosition(x,y,z : double);
var
   v : D3DVECTOR;
begin
     v.x := x;
     v.y := y;
     v.z := z;
     m_Light.dvPosition := v;
end;

function T3dLight.Update : boolean;
begin
     DDResult := m_pILight.SetLight(m_Light);
     if (DDResult <> DD_OK) then
        Result := False
     else
         Result := True;
end;


////////////////////////////////////////////
////////////////////////////////////////////
////////////////////////////////////////////

constructor T3dMaterial.Create(pIEngine : IDirect3d; pIDevice : IDirect3dDevice);
begin
     ZeroMemory(@m_Material,sizeof(m_Material));
     m_Material.dwSize := sizeof(D3DMATERIAL);

     m_hMat := 0;
     m_pIMat := nil;

     m_Material.ambient.r := 1;
     m_Material.Ambient.g := 1;
     m_Material.Ambient.b := 1;
     m_Material.Ambient.a := 1;
     m_Material.Diffuse := m_Material.Ambient;
     m_Material.Specular := m_Material.Ambient;
     m_Material.dwRampSize := 8;

     DDResult := piEngine.CreateMaterial(m_pIMat, nil);
     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create(format ('Cannot create Material!! %x', [DDResult]));

     DdResult := m_pIMat.SetMaterial(m_Material);
     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create(format ('Cannot set Material!! %x', [DDResult]));

     DDResult := m_pIMat.getHandle(pIDevice, m_hMat);
     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create(format ('Cannot get Material handle!! %x', [DDResult]));

end;

function T3dMaterial.GetInterface : IDirect3dMaterial;
begin
     result := m_pIMat;
end;

function T3dMaterial.GetHandle : D3DMATERIALHANDLE;
begin
     result := m_HMat;
end;


function T3dMaterial.SetColor(r,g,b : double) : boolean;
begin
     m_Material.ambient.r := r;
     m_Material.ambient.g := g;
     m_Material.ambient.b := b;
     m_Material.ambient.a := 1;
     m_Material.diffuse.r := r;
     m_Material.diffuse.g := g;
     m_Material.diffuse.b := b;
     m_Material.diffuse.a := 1;
     m_Material.specular.r := 1;
     m_Material.specular.g := 1;
     m_Material.specular.b := 1;
     m_Material.specular.a := 1;
     m_Material.Power := 20.0;

     DdResult := m_pIMat.SetMaterial(m_Material);
     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create(format ('Cannot set Material !! %x', [DDResult]));

     Result := True;
end;

function T3dMaterial.SetShades(n : integer) : boolean;
begin
     m_Material.dwRampSize := n;
     DdResult := m_pIMat.SetMaterial(m_Material);
     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create(format ('Cannot set Material !! %x', [DDResult]));

     Result := True;
end;

function T3DMaterial.SetTextureHandle(h : integer) : boolean;
begin
     m_Material.hTexture := h;
     DDResult := m_pIMat.SetMaterial(m_Material);
     if (DdResult <> DD_OK) then
        RAISE EXCEPTION.create(format ('Cannot set Material !! %x', [DDResult]));

     Result := True;
end;

end.
