unit ttexture;


interface
uses Classes, Forms, SysUtils,  OLE2, Windows , Graphics, DDraw, D3D, D3dTypes, D3DCaps, Direct3d, DirectDraw, dialogs;

// T3DTexture:
// T3DTexture creates, loads and destroys textures [up to 5 textures now but is easy to increase
//  it, just increase the arrays size].
//
// The process to create a texture is the following :
// 1> When initializing set texture names using SetTextureNames(n, fileName)
// 2> When you need to load a texture call LoadTextureSurface(n, bInVideo)
//         bInVideo returns True if the surface is in video memory
// 3> Call GetTextureHandle(n) to return a handle to the texture
// 4> Call ReleaseTexture(n) to discard it

// Most of this code is base on the example tunnel.c and texture.c from the SDK
// I think that the next thing to do is to forget about those arrays and just let
// each instance of the class to represent one texture surface

// To create a T3dTexture object you need a TDDScreen and a TD3D objects.
// TD3D enumerates texture formats that your renderer support and selects one of them
// into TD3D.ThisTextureFormat

// At the moment the texture format remains the same through the execution of the program

type
T3dTexture = class (TObject)
           ptextureSurface : IDirectDrawSurface;
           pTexture : IDIRECT3DTEXTURE;
           scrScreen : TDDScreen;
           devD3D : TD3D;
           TextureHandle : D3DTEXTUREHANDLE;

           function loadSurface(DD : IDirectDraw; fname : string; pformat : DDSurfaceDesc; memoryflag : DWORD) : IDirectDrawSurface;
           function SetTextureHandle : boolean;
public
      constructor Create(scr : TDDScreen; d3d : TD3D);
      destructor Destroy;
      function LoadTextureSurface(fname : string; binVideo : boolean) : boolean;
      function GetTextureHandle : D3DTEXTUREHANDLE;
      procedure ReleaseTexture ;
end;

function LoadPPMHeader(fp : TStream; var width,height : DWORD; var maxgrey : integer) : boolean;
function GetInt(fp : TStream) : integer;
function GetSurfaceDescription(var DDSurfDesc : DDSURFACEDESC;DDSurf : iDIRECTDRAWSURFACE) : HRESULT;
function GetC(f : TStream) : byte;
procedure GetRGB(fp : TStream; bFileType : boolean; var r,g,b :integer; var pal : array of TPaletteEntry);
function LoadBMPHeader(fp : TStream; var width,height : DWORD; var ppe : array of TPaletteEntry; var num_color : integer; var bits : integer) : boolean;

implementation

constructor T3dTexture.Create(scr : TDDScreen; d3d : TD3D);
var
   i : integer;
begin
     scrScreen := scr;
     devD3D := d3d;

     pTextureSurface := nil;
end;

destructor T3dTexture.Destroy;
var
   i : integer;
begin
     if (pTextureSurface <> nil) then
                      ReleaseTexture;
end;

function T3dTexture.LoadTextureSurface(fName : string; binVideo : boolean): boolean;
var
   srctextureSurface : IDIRECTDRAWSURFACE;
   srcTexture : IDIRECT3DTEXTURE;
   ddsd : DDSURFACEDESC;
   pcaps : DWORD;
   ppe : array[0..255] of TPALETTEENTRY;
   DstPalette : IDIRECTDRAWPALETTE;
begin
     if pTextureSurface <> nil then
        ptextureSurface.Release;

     srcTextureSurface := LoadSurface(scrScreen.DDraw, fname, DevD3D.ThisTextureFormat.ddsd, DDSCAPS_SYSTEMMEMORY);
     DDResult := srcTextureSurface.QueryInterface(IID_IDirect3DTexture, srcTexture);
     if (DdResult <> DD_OK) then
        raise exception.create('Cannot get texture interface!');

     DDresult := GetSurfaceDescription(ddsd,srcTextureSurface);
     if (DDResult <> DD_OK) then
        raise exception.create('Cannot get surface description for texture .');

     ddsd.dwFlags := DDSD_CAPS + DDSD_HEIGHT + DDSD_WIDTH + DDSD_PIXELFORMAT;
     ddsd.ddsCaps.dwCaps := DDSCAPS_TEXTURE + DDSCAPS_ALLOCONLOAD;
     if (not ScrScreen.bswRender) then
        ddsd.ddsCaps.dwCaps := ddsd.ddsCaps.dwCaps + DDSCAPS_SYSTEMMEMORY;

     DDResult := scrScreen.DDraw.CreateSurface(ddsd, pTextureSurface, nil);
     if (DDResult <> DD_OK) then
        raise exception.create(format ('Cannot create destination texture surface %x',[DDResult]));

     if (ddsd.ddpfPixelFormat.dwFlags and DDPF_PALETTEINDEXED8) <> 0 then
        pcaps := DDPCAPS_8BIT + DDPCAPS_ALLOW256
     else if (ddsd.ddpfPixelFormat.dwFlags and DDPF_PALETTEINDEXED4) <> 0 then
          pcaps := DDPCAPS_4BIT
     else
         pcaps := 0;

     if (pCaps <> 0) then
        begin
             ZeroMemory(@ppe,sizeof(ppe));

             DDResult := scrScreen.DDraw.CreatePalette(pCaps, @ppe, DstPalette, nil);
             if DDResult <> DD_OK then
                raise exception.create ('Cannot create palette');

             DDResult := ptextureSurface.SetPalette(DstPalette);
             if DDResult <> DD_OK then
                raise exception.create ('Cannot set palette');

             DstPalette.Release;
        end;

     DDResult := ptextureSurface.QueryInterface(IID_IDirect3dTexture, pTexture);
     if DdResult <> DD_OK then
        raise Exception.create('Cannot get 3d texture interface');

     DDResult := ptexture.Load(srcTexture);
     if DDResult <> DD_OK then
        raise exception.create('Cannot load texture');

     srcTexture.release;
     srcTextureSurface.release;

     DDResult := GetSurfaceDescription(ddsd, ptextureSurface);
     if DDResult <> DD_OK then
        raise exception.create('Cannot get surface description');

     if (ddsd.ddsCaps.dwCaps and DDSCAPS_VIDEOMEMORY) <> 0  then
        bInVideo := True
     else
         bInVideo := False;

     Result := SetTextureHandle;
end;

function T3dTexture.SetTextureHandle : boolean;
begin
     DDResult := ptexture.GetHandle(devD3D.D3DDevice, TextureHandle);
     if DdResult <> DD_OK then
        raise exception.create('cannot get handle');

     Result := True;
end;

function T3dTexture.GetTextureHandle : D3DTEXTUREHANDLE;
begin
     Result := TextureHandle;
end;

procedure T3dTexture.ReleaseTexture;
begin
     ptexture.Release;
     ptextureSurface.release;
     texturehandle := 0;
end;

function T3dTexture.loadSurface(DD : IDirectDraw; fname : string; pformat : DDSurfaceDesc; memoryflag : DWORD) : IDirectDrawSurface;
var
   fp : TFileStream;
   dwWidth, dwheight : DWORD;
   i, j : integer;
   fformat : DDSURFACEDESC;
   DDS : IDirectDrawSurface;
   DDPal : IDirectDrawPalette;
   LP : Pinteger;
   SP : ^word;
   PC,BP : ^byte;
   red_shift,green_shift,blue_shift : integer;
   red_scale,green_scale,blue_scale : integer;
   s, pSize,k, r, g, b, color_count, c : integer;
   m : cardinal;
   bQuant : boolean;
   pCaps : DWORD;
   ddsd : DDSURFACEDESC;
   colors : array [0..256] of integer;
   ppe : array [0..256] of TPALETTEENTRY;
   bBitmap : boolean;
   bitsize : integer;
begin

     fp := TFileStream.Create(fname, fmOpenRead);

     bBitmap := False;
     if Pos('.bmp',fname) <> 0 then
        begin
             if (Not LoadBMPHeader(fp, dwWidth, dwHeight, ppe, color_count, bitsize)) then
                begin
                     fp.free;
                     raise exception.create('Cannot parse BMP header');
                end;
             bBitmap := True;
        end
     else if (not LoadPPMHeader(fp, dwWidth, dwHeight, i)) then
        begin
             fp.free;
             raise exception.create('Cannot parse PPM Header');
        end;

     CopyMemory(@fformat, @pformat, sizeof(DDSURFACEDESC));

     if (fformat.ddpfPixelFormat.dwFlags and DDPF_PALETTEINDEXED8)<> 0  then
        begin
             bQuant := TRUE;
             pSize := 256;
             pCaps := DDPCAPS_8BIT + DDPCAPS_ALLOW256;
        end
     else if (fformat.ddpfPixelFormat.dwFlags and DDPF_PALETTEINDEXED4) <> 0 then
         begin
              bQuant := True;
              pSize := 16;
              pcaps := DDPCAPS_4BIT;
         end;

     Copymemory(@ddsd, @fFormat, sizeof(DDSURFACEDESC));
     ddsd.dwSize := sizeof(DDSURFACEDESC);
     ddsd.dwFlags := DDSD_CAPS + DDSD_HEIGHT + DDSD_WIDTH + DDSD_PIXELFORMAT;
     ddsd.ddsCaps.dwCaps := DDSCAPS_TEXTURE + memoryflag;
     ddsd.dwHeight := dwHeight;
     ddsd.dwWidth := dwWidth;

     DDResult := DD.CreateSurface(ddsd,DDS, nil);
     if (DDResult <> DD_OK) then
        begin
             fp.free;
             raise exception.create( Format ('Cannot create surface for texture  %x',[DDResult]));
        end;

     ZeroMemory(@ddsd,sizeof(DDSURFACEDESC));
     ddsd.dwSize := sizeof(DDSURFACEDESC);
     DDResult := DDS.Lock(nil,ddsd,0,0);
     if (DDResult <> DD_OK) then
        begin
             fp.free;
             raise exception.create('Cannot Lock surface');
        end;

     if (not bQuant) then
        begin
             s:=0; m:= fformat.ddpfPixelFormat.dwRBitMask;
             while ((m and 1)=0) do
                   begin
                        Inc(s);
                        m := m shr 1;
                   end;

             red_shift := s;
             red_scale := 255 div (fformat.ddpfPixelFormat.dwRBitMask shr s);

             s:=0; m:= fformat.ddpfPixelFormat.dwGBitMask;
             while ((m and 1)=0) do
                   begin
                        Inc(s);
                        m := m shr 1;
                   end;

             green_shift := s;
             green_scale := 255 div (fformat.ddpfPixelFormat.dwGBitMask shr s);;

             s:=0; m:= fformat.ddpfPixelFormat.dwBBitMask;
             while ((m and 1)=0) do
                   begin
                        Inc(s);
                        m := m shr 1;
                   end;

             blue_shift := s;
             blue_scale := 255 div (fformat.ddpfPixelFormat.dwBBitMask shr s);;

             case (fformat.ddpfPixelFormat.dwRGBBitCount) of
                  32 : begin
                     for j := 0 to dwHeight-1 do
                         begin
                              LP :=  ddsd.lpSurface;
                              k:=0;
                              while (k < (ddsd.lPitch * j)) do
                                    begin
                                         inc(LP);
                                         inc(k);
                                    end;

                              for i:=0 to dwWidth-1 do
                                  begin
                                       GetRGB(fp, bBitmap, r,g,b, ppe);

                                       LP^ := (r shl red_shift) + (g shl green_shift) + (b shl blue_shift);
                                       Inc(LP);
                                  end;
                         end;
                     end;
                  16: begin
                     for j := 0 to dwHeight-1 do
                         begin
                              SP := ddsd.lpSurface;
                              k:=0;
                              while (k < (ddsd.lPitch * j)) do
                                    begin
                                         inc(SP);
                                         inc(k);
                                    end;
                              //SP := SP + (ddsd.lPitch * j);

                              for i:=0 to dwWidth-1 do
                                  begin
                                       GetRGB(fp, bBitmap, r,g,b , ppe);
                                       SP^ := (r shl red_shift) + (g shl green_shift) + (b shl blue_shift);
                                       Inc(SP);
                                  end;
                         end;
                     end;
                  8: begin
                     for j := 0 to dwHeight-1 do
                         begin
                              BP := ddsd.lpSurface;
                              k:=0;
                              while (k < (ddsd.lPitch * j)) do
                                    begin
                                         inc(BP);
                                         inc(k);
                                    end;
                              //BP := BP + (ddsd.lPitch * j);

                              for i:=0 to dwWidth-1 do
                                  begin
                                       GetRGB(fp, bBitmap, r,g,b, ppe);

                                       BP^ := (r shl red_shift) + (g shl green_shift) + (b shl blue_shift);
                                       Inc(BP);
                                  end;
                         end;
                     end;
                  else
                      begin
                           DDS.Unlock(nil);
                           fp.free;
                           DDS.Release;
                           Raise Exception.Create('Unknown pixel format for texture.');
                      end;
             end;

             DDS.Unlock(nil);
             fp.free;
             Result := DDS;
             exit;
        end;

     if (not bBitmap) then
        begin
             color_count :=0;
             ZeroMemory(@colors,sizeof(colors));
             for j:=0 to dwHeight-1 do
                 begin
                 PC := ddsd.lpSurface;
                 k :=0;
                 while (k < (ddsd.lPitch * j)) do
                    begin
                         Inc(PC);
                         inc(k);
                    end;
              //PC := PC + (ddsd.lPitch * j);
                 for i:=0 to dwWidth-1 do
                  begin
                       r := getc(fp);
                       g := getc(fp);
                       b := getc(fp);
                       c := (r shl 16) + (g shl 8) + b;
                       k := 0; //(((r * 1000) + (g * 100) + b)*255) div 280755;
                       while (k < color_count) do
                          begin
                             if (c = colors[k]) then break;
                             inc(k);
                          end;

                       if (k = color_count) then
                          begin
                               inc(color_count);

                               if (color_count > psize) then
                                  begin
                                       Dec(color_count);
                                       k := color_count - 1;
                                  end;

                               colors[k] := c;
                          end;

                       if (psize = 16) then
                         begin
                          if ((i and 1) = 0) then
                             PC^ := k and $f
                          else
                              begin
                                   PC^ := PC^ + ((k and $f) shl 4);
                                   Inc(PC);
                              end;
                         end
                       else
                           begin
                                PC^ := byte (k);
                                Inc(PC);
                           end;

                  end;
                 end;
              if (color_count > psize) then
                 begin
                      DDS.Release;
                      Raise Exception.create('Palette burst');
                 end;

              Zeromemory(@ppe, sizeof(TPALETTEENTRY) * 256);
              for i:=0 to color_count do
                     begin
                          ppe[i].peRed := (colors[i] shr 16);
                          ppe[i].peGreen := (colors[i] shr 8);
                          ppe[i].peBlue := (colors[i] and 255);
                          ppe[i].peFlags := D3DPAL_RESERVED;
                     end;
        end
     else
         begin
           // Easy bitmap loading
              PC := ddsd.lpSurface;

              fp.Read(PC^, round(dwHeight * (((dwWidth * bitSize) + 31) / 32 * 4)));
              for i:=0 to color_count-1 do
                     begin
                          ppe[i].peFlags := D3DPAL_RESERVED;
                     end;
              ppe[0].peFlags := D3DPAL_READONLY;
              ppe[255].peFlags := D3DPAL_READONLY;

         end;

     fp.free;
     DDS.Unlock(nil);

     DDResult := DD.CreatePalette(DDPCAPS_INITIALIZE + pcaps, @ppe, DDPal, nil);
     if (DDResult <> DD_OK) then
        begin
             DDS.Release;
             Raise Exception.Create('Cannot create D3D palette ');
        end;

     DDResult := DDS.SetPalette(DDPAL);
     if (DDResult <> DD_OK) then
        begin
             DDS.Release;
             DDpal.release;
             Raise Exception.Create('Cannot set palette');
        end;

     DDPal.Release;

     Result := DDS;
end;


function GetC(f : TStream) : byte;
var
   b : byte;
begin
     f.ReadBuffer(b,1);
     result := b;
end;

procedure GetRGB(fp : TStream; bFileType : boolean; var r,g,b :integer; var pal : array of TPaletteEntry);
var
   k : integer;
begin
     if not bFileType then
        begin
             r := GetC(fp);
             g := GetC(fp);
             b := GetC(fp);
        end
     else
         begin
              k := GetC(fp);
              r := pal[k].peRed;
              g := pal[k].peGreen;
              b := pal[k].peBlue;
         end;
end;


function GetInt(fp : TStream) : integer;
var
   cInt : integer;
   c : byte;
begin
  repeat
     c := getc(fp);
  until (C <> ord(' '))and(c <> 10);

  cint := 0;

  repeat                           //0 ascii
        cint := (cint * 10) + (c - ord('0'));
        c := getc(fp);
  until (c < ord('0')) or (c > ord('9'));//   <--- use ascii codes

  result := cint;
end;

function LoadPPMHeader(fp : TStream; var width,height : DWORD; var maxgrey : integer) : boolean;
var
   magic : array [0..2] of byte;
   c : byte;
begin
     magic[0] := GetC(fp);
     magic[1] := GetC(fp);
     Getc(fp);
//     if (magic[0] <> Check out
     if (Getc(fp) = ord('#')) then
          begin
             while (Getc(fp) <> 10) do;
          end
     else
         fp.seek(-1, soFromCurrent);

     width := GetInt(fp);
     height := getInt(fp);
     maxgrey := getint(fp);

     repeat
           c := getc(fp);
     until (c <> ord(' '));//and(c <> ord(''))and(c<>ord('\n'))and(c<>ord('\r'));  // or other options

     fp.seek(-1, soFromCurrent);

     Result := True;
end;

function GetSurfaceDescription(var DDSurfDesc : DDSURFACEDESC;DDSurf : iDIRECTDRAWSURFACE) : HRESULT;
begin
    ZeroMemory(@DDSurfDesc,sizeof(DDSURFACEDESC));
    DDSurfDesc.dwSize := sizeof(DDSURFACEDESC);
    result := DDSurf.GetSurfaceDesc(DDSurfDesc);
end;

function LoadBMPHeader(fp : TStream; var width,height : DWORD;  var ppe : array of TPaletteEntry; var num_color : integer; var bits : integer) : boolean;
var
   bm : TBITMAPINFOHEADER;
   bf : TBITMAPFILEHEADER;
begin
     fp.ReadBuffer(BF, sizeof(TBITMAPFILEHEADER));
     fp.ReadBuffer(BM, sizeof(TBITMAPINFOHEADER));

     width := bm.biWidth;
     height := bm.biHeight;
     bits := bm.biBitCount;

     if (bm.bibitCount = 8) then
        num_color := 256
     else if (bm.biBitCount = 4) then
          num_color := 16;

     fp.Read(ppe, sizeof(TRGBTRIPLE) * num_color);

     fp.Seek(bf.bfOffBits + sizeof(TBITMAPFILEHEADER) - 1, soFromBeginning);

     Result := True;
end;



end.
